package com.mjitech.api

import com.fasterxml.jackson.databind.node.ObjectNode
import com.timzaak.jackson.KtorJacksonWrapper
import config.AppConfig
import io.ktor.client.HttpClient
import io.ktor.client.features.HttpTimeout
import io.ktor.client.features.timeout
import io.ktor.client.request.HttpRequestBuilder
import io.ktor.client.request.forms.FormDataContent
import io.ktor.client.request.request
import io.ktor.client.statement.HttpResponse
import io.ktor.client.statement.readText
import io.ktor.http.HttpMethod
import io.ktor.http.Parameters
import io.ktor.http.isSuccess
import io.ktor.http.parametersOf
import io.ktor.http.plus
import kotlinx.coroutines.CoroutineScope
import java.util.Base64
import kotlin.coroutines.CoroutineContext

interface API : CoroutineScope {
    fun HttpRequestBuilder.addSecret()

    // for http forward
    suspend fun post(uri: String, body: Parameters): HttpResponse
    suspend fun posPost(body: Parameters): HttpResponse
}

class BaseAPIImpl(
        private val appConfig: AppConfig,
        private val httpClient: HttpClient,
        override val coroutineContext: CoroutineContext
) : API {
    private val ServerCertHeaderName = "mjitech-machine-cert"
    private val Cert = Base64.getEncoder().encodeToString(appConfig.serverCert.toByteArray())

    override fun HttpRequestBuilder.addSecret() {
        headers.append(ServerCertHeaderName, Cert)
    }

    override suspend fun post(uri: String, body: Parameters): HttpResponse {
        val url = "${appConfig.remoteServer}$uri"
//        val partHeaders = HeadersBuilder().apply {
//            append(HttpHeaders.ContentDisposition, "form-data; name=storeId")
//        }

        return httpClient.request(url) {
            method = HttpMethod.Post
            // contentType(ContentType.Application.Json)
            addSecret()
            val form = FormDataContent(parametersOf("storeId", appConfig.storeId) + body)
            this.body = form
        }
    }

    override suspend fun posPost(body: Parameters): HttpResponse {
        return httpClient.request(appConfig.cardPayUrl) {
            method = HttpMethod.Post
            //TODO:如何去处理连接pos机10秒出现的timeOut异常
            timeout {
                requestTimeoutMillis = 1*60*1000
                socketTimeoutMillis =  1*60*1000
                connectTimeoutMillis = 10* 1000
            }
            val form = FormDataContent(parametersOf("storeId", appConfig.storeId) + body)
            this.body = form
        }
    }
}

suspend fun HttpResponse.jsonResult(): Result<ObjectNode> {
    return runCatching {
        val str = readText(Charsets.UTF_8)
        KtorJacksonWrapper.jacksonMapper.readTree(str) as ObjectNode
    }
}

suspend inline fun <reified T> HttpResponse.result(dataPath: String?): Result<T> {
    if (status.isSuccess()) {
        val str = readText(Charsets.UTF_8)
        val json = KtorJacksonWrapper.jacksonMapper.readTree(str) as ObjectNode
        if (json["is_succ"].asBoolean()) {
            if (dataPath == null) {
                return Result.success(
                    KtorJacksonWrapper.jacksonMapper.convertValue(json, T::class.java)
                )
            } else {
                return Result.success(
                    KtorJacksonWrapper.jacksonMapper.convertValue(json[dataPath], T::class.java)
                )
            }
        }
    }
    return Result.failure(Exception("server error: ${readText((Charsets.UTF_8))}"))
}

suspend inline fun <reified T> HttpResponse.resultWithB(): Result<Pair<Boolean, T>> {
    return resultWithB(null)
}
suspend inline fun <reified T> HttpResponse.resultWithB(dataPath: String?): Result<Pair<Boolean, T>> {
    if (status.isSuccess()) {
        val str = readText(Charsets.UTF_8)
        val json = KtorJacksonWrapper.jacksonMapper.readTree(str) as ObjectNode
        val isSuccess = json["is_succ"].asBoolean()

        if (dataPath == null) {
            return Result.success(
                isSuccess to KtorJacksonWrapper.jacksonMapper.convertValue(json, T::class.java)
            )
        } else {
            return Result.success(
                isSuccess to KtorJacksonWrapper.jacksonMapper.convertValue(json[dataPath], T::class.java)
            )
        }
    }
    return Result.failure(Exception("server error: ${readText((Charsets.UTF_8))}"))
}

suspend inline fun HttpResponse.result(): Result<Unit> {
    if (status.isSuccess()) {
        val str = readText(Charsets.UTF_8)
        val json = KtorJacksonWrapper.jacksonMapper.readTree(str) as ObjectNode
        if (json["is_succ"].asBoolean()) {
            return Result.success(Unit)
        } else {
            return Result.failure(Exception("server error: $str"))
        }
    } else {
        return Result.failure(Exception("server error: ${readText((Charsets.UTF_8))}"))
    }
}
// abstract class API(private val api: BaseAPIImpl) : APIInterface by api
