package com.mjitech.service

import com.mjitech.iot.DeviceStatus
import com.mjitech.service.log.DownLog
import com.mjitech.service.state.StateManager
import com.mjitech.service.state.StateTransition
import com.mjitech.service.state.entity.Event
import com.mjitech.service.state.entity.SideEffect
import com.mjitech.service.state.entity.State
import com.mjitech.third.down.DownClient
import com.mjitech.third.down.entity.*
import com.mjitech.third.down.entity.server.ACKResponse
import com.mjitech.third.down.entity.server.HeartBeatResponse
import kotlinx.coroutines.channels.consumeEach
import kotlinx.coroutines.delay
import kotlinx.coroutines.isActive
import kotlinx.coroutines.launch
import mu.KotlinLogging
import org.joda.time.DateTime
import kotlin.coroutines.CoroutineContext

class DownClientService(
    private val downClient: DownClient,
    private val stateManager: StateManager,
    override val coroutineContext: CoroutineContext
) : Service {

    private val logger = KotlinLogging.logger { }
    private var isTestDownMachineLost = false

    init {
        launch {
            listenDownMessage()
        }
        launch {
            checkDownServer()
        }
        launch {
            sendHeartBeat()
        }
    }

    private fun getDelayTime(): Long {
        // 访问下位机时间间隔
        // 下位机设计问题，拿货的时候，必须快速取状态
        if (stateManager.state is State.DoBusiness) {
            return 100L
        }
        return 1000L
    }

    private fun isDownMachineAlive(): Boolean {
        return DateTime.now().minusSeconds(21).isBefore(receiveMessageTime)
    }



    suspend fun sendHeartBeat() {
        launch {
            while (isActive) {
                delay(getDelayTime())
                downClient.sendCommand(HeartBeatRequest)
            }
        }
    }

    override suspend fun consume(transition: StateTransition) {
        if(transition.toState.desc().id == DeviceStatus.Free) {
            downClient.sendCommand(ResetSCM)
        }
        transition.sideEffect.let { sideEffect ->
            when (sideEffect) {
                is SideEffect.OpenDoorInRepair -> {
                    when (sideEffect.doorId) {
                        State.Inspection.OpenDoor.ElectronDoor -> {
                            downClient.sendCommand(ElectronDoor(ElectronDoor.OpenDoor))
                        }
                        State.Inspection.OpenDoor.ADoor, State.Inspection.OpenDoor.BDoor -> {
                            val area = if (sideEffect.doorId == State.Inspection.OpenDoor.ADoor) {
                                StorageDoor.AreaA
                            } else {
                                StorageDoor.AreaB
                            }
                            downClient.sendCommand(StorageDoor(StorageDoor.OpenDoor, area))
                        }
                        else -> {
                        }
                    }
                }
                is SideEffect.TestDownMachineLost -> {
                    this.isTestDownMachineLost = sideEffect.isLost
                }
                else -> {
                }
            }
        }
    }

    private var receiveMessageTime = DateTime.now()

    private suspend fun checkDownServer() {
        while (isActive) {
            delay(1 * 1000)
            if (!isDownMachineAlive() && stateManager.state !is State.MachineError) {
                stateManager.handleEvent(Event.DownMachineLost)
            }
        }
    }

    // 心跳检测 下位机器是否 OK
    private suspend fun listenDownMessage() {
        logger.debug { "begin to listen down message" }
        downClient.downClientChannel.channel.openSubscription().consumeEach { command: Command ->
            // 下位机响应了重新开始心跳
            // 更新下位机器发送指令时间
            receiveMessageTime = DateTime.now()
            when (command) {
                is HeartBeatResponse -> {
                    stateManager.setDownClientStatus(command)
                    // TODO: 异常 DownHeartBeat 没有进入到状态机 逻辑里面去
                    val downMachineIsSafe = command.isSafe()
                    if (downMachineIsSafe && stateManager.state is State.Init) {
                        // 初始化判定
                        stateManager.handleEvent(Event.InitEvent.DownClientInitSuccess)
                    } else if(!isTestDownMachineLost && downMachineIsSafe && stateManager.state is State.MachineError.DownMachineLost){
                        //下位机失联后重新检测到心跳
                        stateManager.handleEvent(Event.SetState(DeviceStatus.Free))
                    }else if (!downMachineIsSafe) {
                        stateManager.handleEvent(
                            Event.DownMachineIsNotSafe(
                                command.leftHorseBlock,
                                command.rightHorseBlock,
                                command.shipmentWarehouse
                            )
                        )
                        //打印下位机是否安全日志
                        DownLog.DownIsSafeLog(command)
                    }


                }
                is ACKResponse -> {
                    // 接口设计的有问题, 先这么 SB 的处理一下
                    stateManager.setDownClientStatus(
                        HeartBeatResponse.fromACKResponse(command)
                    )

                    val downMachineIsSafe = command.isSafe()
                    if (downMachineIsSafe && stateManager.state is State.Init) {
                        // 初始化判定
                        stateManager.handleEvent(Event.InitEvent.DownClientInitSuccess)
                    } else if(!isTestDownMachineLost && downMachineIsSafe && stateManager.state is State.MachineError.DownMachineLost){
                        //下位机失联后重新检测到心跳
                        stateManager.handleEvent(Event.SetState(DeviceStatus.Free))
                    }else if (!downMachineIsSafe) {
                        stateManager.handleEvent(
                                Event.DownMachineIsNotSafe(
                                        command.leftHorseBlock,
                                        command.rightHorseBlock,
                                        command.shipmentWarehouse
                                )
                        )
                        //打印下位机是否安全日志
                        DownLog.DownIsSafeLog(command)
                    }
                }
                else -> {
                }
            }
        }
    }

}
