package com.mjitech.service

import com.fasterxml.jackson.databind.ObjectMapper
import com.mjitech.api.OrderAPI
import com.mjitech.api.SkuAPI
import com.mjitech.api.model.FailTaking
import com.mjitech.api.model.OrderStatus
import com.mjitech.api.model.SKU
import com.mjitech.api.model.Order
import com.mjitech.service.state.StateManager
import com.mjitech.service.state.entity.Event
import com.mjitech.service.state.entity.SideEffect
import com.mjitech.service.state.entity.State
import com.mjitech.third.down.DownClient
import com.mjitech.third.down.entity.*
import com.mjitech.third.down.entity.server.MachineErrorResponse
import com.mjitech.third.down.entity.server.OutGoodDoorStatus
import com.mjitech.third.down.entity.server.OutGoodStatus
import com.mjitech.third.down.entity.server.OutGoodStatus.Companion.Status
import com.mjitech.third.scancode.ScanCodeClient
import com.tinder.StateMachine
import kotlinx.coroutines.channels.consumeEach
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import mu.KotlinLogging
import kotlin.coroutines.CoroutineContext

class ShopService(
    private val orderAPI: OrderAPI,
    private val skuAPI: SkuAPI,
    private val stateManager: StateManager,
    private val downClient: DownClient,
    private val scanCodeClient: ScanCodeClient,
    private val objectMapper: ObjectMapper,
    override val coroutineContext: CoroutineContext
) : Service {
    private val logger = KotlinLogging.logger {}

    init {
        logger.info { "init shopService" }
        handleScanCode()
        handleDownMessage()
    }

    suspend fun updateSkuState(order: Order, batchStep: Int, step: Int, status: Int) {
        val sku = order.getSKUByStep(batchStep,step)!!
        skuAPI.setOutBatchSkuStatus(
            sku.outBatchId,
            sku.outBatchSkuId,
            order.orderNumber,
            status
        )
    }

    suspend fun getExchangeSku(order: Order, batchStep: Int, step: Int) {
        val sku = order.getSKUByStep(batchStep,step)!!
        val result =
            orderAPI.getExchangeSku(sku.outBatchId, sku.outBatchSkuId, order.orderNumber, sku.skuId)
        result.onSuccess { (isSuccess, res) ->
            if (isSuccess) {
                if (res.originalSku.status == SKU.Companion.Status.READY &&
                    res.sku!!.status == SKU.Companion.Status.PAID
                ) {
                    order.replaceSku(res.removePresent,batchStep)
                    logger.info { "take new good" }
                    //直接封装换货的sku为batch,发送给下位机取货
                    downClient.sendCommand(OutGoods.fromSKU(sku))
                    // 修改错误sku状态
                    launch {
                        updateSkuState(order,batchStep, step, SKU.Companion.Status.READY)
                    }
                }
            } else {
                if (res.originalSku.status == SKU.Companion.Status.REFOUND ||
                    res.originalSku.status == SKU.Companion.Status.REFUND_FAIL
                ) {
                    logger.info { "take new good failure" }
                    order.replaceSku(res.removePresent,batchStep)
                    launch {
                        updateSkuState(order, batchStep, step, res.originalSku.status)
                    }
                    stateManager.handleEvent(Event.ChangeSkuFailure)
                } else {
                    delay(1000)
                    getExchangeSku(order, batchStep, step)
                }
            }
        }
    }

    override suspend fun consume(transition: StateMachine.Transition.Valid<State, Event, SideEffect>) {
        transition.sideEffect?.let { effect ->
            when (effect) {
                is SideEffect.TakeBatch -> {
                    downClient.sendCommand(ResetSCM)
                    logger.info { "begin to take batch, batchStep:${effect.batchStep}" }
                    if(effect.batchStep == 0){
                        launch {
                            orderAPI.updateOrderStatus(effect.order.orderNumber, OrderStatus.TakingOrder)
                        }
                    }
                    if(effect.order.batches.size == effect.batchStep){
                        logger.info { "update order to finish" }
                        launch {
                            orderAPI.updateOrderStatus(effect.order.orderNumber, OrderStatus.FinishOrder)
                        }
                    }else{
                        //取下一个batch
                        delay(100)
                        logger.info { "take next batch batchStep:${effect.batchStep}" }
                        val command = OutGoods.fromOrderBatch(effect.order.getBatchByStep(effect.batchStep)!!)
                        downClient.sendCommand(command)
                    }

                }
                is SideEffect.TakeGood -> {
                    // 拿一件货成功
                    logger.info("begin to take good, step:${effect.step} success")
                    launch {
                        updateSkuState(effect.order, effect.batchStep,effect.step - 1, SKU.Companion.Status.FinishTaking)
                    }
                }
                is SideEffect.SetStatusTaking ->{
                    updateSkuState(effect.order,effect.batchStep,effect.step,SKU.Companion.Status.Taking)
                }
                is SideEffect.OpenDoor -> {
                    logger.info { "batch拿货完成: ${effect.order.batches[effect.batchStep].id}, 准备开门" }
                    launch {
                        // last get SKU
                        updateSkuState(effect.order,effect.batchStep, effect.order.batches[effect.batchStep].skus.size - 1, SKU.Companion.Status.FinishTaking)
                    }
                    downClient.sendCommand(OutGoodDoor)

                }
                is SideEffect.TakeGoodFailure -> {
                    logger.info("take good failure order:${effect.order.id}, takingWhich: ${effect.failureStep}")
                    launch {
                        updateSkuState(effect.order,effect.batchStep, effect.failureStep, SKU.Companion.Status.Missing)
                    }
//                val (batch, sku) = effect.order.getOrderBatchAndSKUByStep(effect.failureStep)!!
                    val sku = effect.order.getSKUByStep(effect.batchStep,effect.failureStep)!!
                    val failTaking = skuAPI.notifySKUMiss(sku.outBatchId, sku.outBatchSkuId, effect.order.orderNumber)
                     failTaking.onFailure {
                         logger.info { "获取换货商品失败" }
                     }
                    failTaking.onSuccess { (isSuccess, res) ->
                        if (isSuccess) {
                            logger.info { "success fail_get_sku" }
                            // 开始换货等候7寸屏选择换货或退款
                            stateManager.handleEvent(
                                Event.ChangeSku(
                                    effect.order, effect.failureStep,
                                    FailTaking(
                                        effect.order.orderNumber,
                                        true,
                                        sku,
                                        res.skus
                                    )
                                )
                            )
                        }
                    }
                }
                is SideEffect.ChangeSku -> {
                    logger.info { "changeSku batchStep:${effect.batchStep},step:${effect.step}" }
                    getExchangeSku(effect.order, effect.batchStep, effect.step)
                }
                is SideEffect.MachineError -> {
                    // 机器故障
                    logger.info { "machineError-------" }
                    val sku = effect.order.getSKUByStep(effect.batchStep,effect.failureStep)!!
                    val isOpenDoor = stateManager.state is State.DoBusiness.OpenGoodDoor
                    // 对返回不做处理
                    val result =
                        orderAPI.reportError(0, "机器故障", isOpenDoor, sku.outBatchId, sku.outBatchSkuId)
                    logger.info { "upload_error body : $result" }
                    delay(60000)
                    stateManager.handleEvent(Event.UpdateMachinePc)
                }
                is SideEffect.TestTakingSku -> {
                    logger.info { "test down take good " }
                    downClient.sendCommand(
                            OutGoods(
                                    listOf(OutGood(
                                            effect.shelfNumber.toByte(),
                                            effect.x,
                                            effect.y,
                                            effect.x1,
                                            effect.y1,
                                            1.toByte(),
                                            1.toByte()
                                    ))
                            )

                    )
                }
                else -> {
                }
            }
        }
    }

    private fun handleDownMessage() {
        launch {
            downClient.downClientChannel.channel.openSubscription().consumeEach {
                when (it) {
                    is OutGoodStatus -> {
                        logger.info { "outgoodstatus consume ${it.status}" }

                        when (it.status) {
                            Status.IS_TAKING -> {
                                stateManager.handleEvent(Event.TakingSucc)
                            }
                            Status.PRODUCT_REMAIN -> {
                                stateManager.handleEvent(Event.ProductRemain)
                            }
                            Status.TakeGoodMiss -> {
                                stateManager.handleEvent(Event.TakeGoodFailure(it.status,it.takingWhich.toInt()))
                            }
                            Status.TakeSingleGood -> {
                                stateManager.handleEvent(Event.TakeGoodOk)
                            }
                            Status.READY -> {
                                stateManager.handleEvent(Event.DownPrepareForNext)
                            }
                            Status.MachineError -> {
                                stateManager.handleEvent(Event.MachineError(it.status))
                                downClient.sendCommand(MachineError)
                            }
                            else -> {
                                logger.debug { "do not handle OutGoodStatus:${it.status.toInt()}" }
                            }
                        }
                    }
                    is OutGoodDoorStatus -> {
                        logger.info { "outGoodDoorStatus consume ${it.status}" }
                        when (it.status) {
                            OutGoodDoorStatus.opening -> {
                                stateManager.handleEvent(Event.OpeningDoor)
                            }
                            OutGoodDoorStatus.openSuccess -> {
                                stateManager.handleEvent(Event.OpenSuccessDoor)
                            }
                        }
                    }
                    is MachineErrorResponse -> {
                        orderAPI.reportDownError(it.superType.toInt(),it.subType.toInt())
                    }

                }
            }
        }
    }

    private var localTakingNumberCache = listOf<String>()

    // TODO: ???
    private fun checkLocalTakingNumberCache(number: String): Boolean {
        val MAX_CACHE_LENGTH = 20
        if (number != "" && localTakingNumberCache.indexOf(number) == -1) {
            if (localTakingNumberCache.size >= MAX_CACHE_LENGTH) {
                localTakingNumberCache.toMutableList().removeAt(0)
            }
            localTakingNumberCache.toMutableList().add(number)
            return true
        } else {
            return false
        }
    }


    private fun handleScanCode() {
        launch {
            scanCodeClient.scanCodeChannel.channel.consumeEach { text ->
                logger.info { "scanCode stateManager.state..... ${stateManager.state}" }
                // TODO: LocalTakingNumberCache 逻辑，是否需要？
                // 取货码多次扫入，不重复取货
                if (checkLocalTakingNumberCache(text)) {
                    logger.info { checkLocalTakingNumberCache(text) }
                    if (stateManager.state is State.Free) {
                        orderAPI.getOrderDetailByTokenId(text).map { data ->
                            if (data["is_succ"].asBoolean()) {
                                // TODO: 如果订单货物已被拿，该怎么做？
                                 var order = objectMapper.convertValue(data["order"], Order::class.java)
                                logger.info { "takingNumber:${text} to order:${order} " }
                                if (order.batches.isNotEmpty()) {
                                    stateManager.handleEvent(Event.ScanCodeEvent(order, text))
                                } else {
                                    logger.warn { "scanCode:$text has empty order" }
                                }
                            } else {
                                logger.warn { "scanCode is error" }
                                if (data["error_code"].asInt() == -10) {
                                    // webResponseCache.set(data)
                                    stateManager.setErrorOrderInfo(data)
                                    // 防止处理的过于频繁,可有可无
                                    delay(1000)
                                } else {
                                    downClient.sendCommand(ResetSCM)
                                }
                            }
                        }
                    } else {
                        logger.debug { "scan code:$text, does not trigger" }
                    }
                }
            }
        }
    }
}
