package com.mjitech.service.mqtt

import com.fasterxml.jackson.databind.ObjectMapper
import com.mjitech.iot.DoorId
import com.mjitech.service.Service
import com.mjitech.service.mqtt.command.*
import com.mjitech.service.state.DoorState
import com.mjitech.service.state.StateManager
import com.mjitech.service.state.entity.Event
import com.mjitech.service.state.entity.SideEffect
import com.mjitech.service.state.entity.State
import com.mjitech.third.down.DownClient
import com.mjitech.third.down.entity.*
import com.mjitech.third.down.entity.server.ACKResponse
import com.mjitech.third.down.entity.server.HeartBeatResponse
import com.mjitech.third.down.entity.server.HeartBeatResponse.Companion.Safe
import com.timzaak.docker.DockerClient
import com.timzaak.jackson.jsObject
import com.timzaak.net.LocalNetInfo
import com.timzaak.ucloud.mqtt.UIotFileClient
import com.timzaak.ucloud.mqtt.UMqttClient
import com.timzaak.ucloud.mqtt.entity.MqttState
import com.timzaak.ucloud.mqtt.entity.PropertyValue
import com.timzaak.ucloud.mqtt.entity.ResponseCommand
import com.timzaak.ws.ConnectionManager
import com.tinder.StateMachine
import config.AppConfig
import io.ktor.http.cio.websocket.Frame
import jnr.ffi.annotations.Out
import kotlinx.coroutines.channels.consumeEach
import kotlinx.coroutines.launch
import mu.KotlinLogging
import java.nio.file.Files
import java.nio.file.Paths
import kotlin.coroutines.CoroutineContext

class CloudMqttService(
    private val uMqttClient: UMqttClient,
    private val uIotFileClient: UIotFileClient,
    private val objectMapper: ObjectMapper,
    private val dockerClient: DockerClient,
    private val stateManager: StateManager,
    private val downClient: DownClient,
    private val appConfig: AppConfig,
    private val connectionManager: ConnectionManager,
    override val coroutineContext: CoroutineContext
) : Service {
    private val logger = KotlinLogging.logger { }

    init {
        launch {
            listenMqttStateMessage()
        }
        launch {
            listenDownClientMessage()
        }
        launch {
            listenCloudMessage()
        }
    }

    private fun doorIdToCommand(doorId: Int): CommandRequest? {
        return when (doorId) {
            State.Inspection.OpenDoor.ElectronDoor -> {
                ElectronDoor(ElectronDoor.OpenDoor)
            }
            State.Inspection.OpenDoor.ADoor -> {
                StorageDoor(StorageDoor.OpenDoor, StorageDoor.AreaA)
            }
            State.Inspection.OpenDoor.BDoor -> {
                StorageDoor(StorageDoor.OpenDoor, StorageDoor.AreaB)
            }
//            State.Inspection.OpenDoor.OutGoodDoor -> {
//                // do not implement
//            }
            else -> {
                logger.warn { "do not implement door command:$doorId" }
                null
            }
        }
    }

    override suspend fun consume(transition: StateMachine.Transition.Valid<State, Event, SideEffect>) {
        if (uMqttClient.client.isConnected) {

            if (transition.fromState is State.Init && transition.toState is State.Free) {
                // 初始化
                uMqttClient.setServerInfoWhenInitToFree()
            } else if (transition.fromState.desc().id != transition.toState.desc().id) {
                // 状态变化上报
                uMqttClient.setState(transition.toState.desc().id)
            }

            transition.sideEffect?.let { sideEffect ->
                when (sideEffect) {
                    is SideEffect.OpenDoorInRepair -> {
                        doorIdToCommand(sideEffect.doorId)?.let {
                            downClient.sendCommand(it)
                        }
                    }
                    else -> {
                    }
                }
            }
        } else {
            logger.warn { "mqtt does not connect" }
        }
    }

    // 上报门状态
    private suspend fun listenDownClientMessage() {
        downClient.downClientChannel.channel.openSubscription().consumeEach { command ->
            when (command) {
                is HeartBeatResponse -> {
                    //上传A,B门,电柜门状态
                    stateManager.setNewDoorState(DoorState(command.leftDoor == Safe, command.rightDoor == Safe, command.electronDoor == Safe))?.let {
                        uMqttClient.setProperty(it)
                    }
                }
                is ACKResponse -> {
                    //上传A,B门,电柜门状态
                    stateManager.setNewDoorState(DoorState(command.leftDoor == Safe, command.rightDoor == Safe, command.electronDoor == Safe))?.let {
                        uMqttClient.setProperty(it)
                    }
                }
            }
        }
    }

    private suspend fun listenMqttStateMessage() {
        uMqttClient.mqttStateChannel.channel.openSubscription().consumeEach { state ->
            if (state is MqttState.Connected) {
                stateManager.handleEvent(Event.InitEvent.MqttClientInitSuccess)
            }
        }
    }

    private suspend fun listenCloudMessage() {
        uMqttClient.mqttMessageChannel.channel.openSubscription().consumeEach { (topic, message) ->
            logger.debug {
                "from cloud, topic:$topic, message: ${objectMapper.readTree(message.payload).toPrettyString()}"
            }
            val json = objectMapper.readTree(message.payload)
            when (topic) {
                uMqttClient.CommandTopic -> {
                    when (json.getIdentifier()) {
                        SetStatus.identifier -> {
                            val da = objectMapper.convertValue(json, DataBody::class.java).Input
                            val data = objectMapper.convertValue(da, Event.SetState::class.java)
                            logger.info { "SetStatus data : $data" }
                            stateManager.handleEvent(data)
                        }
//                        UploadLog.identifier -> {
//                            // TODO: 日志上传
//                            val logName = "TODO.log"
//                            val tmpName = "TODO-bak.log"
//                            val logFile = Paths.get(appConfig.logPath, logName)
//                            val tmpPath = Paths.get(appConfig.logPath, tmpName)
//                            Files.copy(logFile, tmpPath)
//                            uIotFileClient.postFile(tmpPath.toFile(), tmpName)
//                        }
                        OpenDoor.identifier -> {
                            val da = objectMapper.convertValue(json, DataBody::class.java).Input
                            val data = objectMapper.convertValue(da, Event.OpenDoorInRepair::class.java)
                            logger.info { "OpenDoor data : $data" }
                            if (data.doorId == DoorId.ADoor || data.doorId == DoorId.BDoor) {
                                val area = if (data.doorId == DoorId.ADoor) {
                                    "A"
                                } else {
                                    "B"
                                }
                                connectionManager.broadcastAll(
                                    Frame.Text(
                                        jsObject {
                                            "type" /= "SUCCESS_OPENDOOR"
                                            "info" /= jsObject {
                                                "userinfo" /= jsObject {
                                                    "employeeId" /= data.employeeId
                                                }
                                                "area" /= area
                                                "maxTime" /= 300
                                            }
                                        }.toString()
                                    )
                                )
                            }
                            stateManager.handleEvent(data)
                        }
                        ReloadBrowser.identifier -> {
                            logger.info { "ReloadBrowser" }
                            stateManager.setWebNeedReload()
                        }
                        ReloadServer.identifier -> {
                            logger.info { "ReloadServer" }
                            //重启上位机Docker
                            dockerClient.restartContainer()
                        }
                        NewSkuPass.identifier -> {
                            val da = objectMapper.convertValue(json, DataBody::class.java).Input
                            val data = objectMapper.convertValue(da, Event.NewSkuPass::class.java)
                            logger.info { "NewSkuPass data : $data" }
                            stateManager.handleEvent(data)
                        }
                        OpenGoodDoor.identifier -> {
                            logger.info { "OpenGoodDoor" }
                            downClient.sendCommand(OutGoodDoor)
                        }
                        /*
                        GetTemperature.identifier -> {
                            downClient.sendCommand(QueryTemperature)
                            return
                        }*/
                        else -> {
                            logger.warn { "do not handle command from mqtt: $json" }
                        }
                    }
                    // 不管如何, 都直接返回
                    uMqttClient.respCommand(ResponseCommand.baseResp(json))
                }
                else -> {
                    logger.info { "do not handle topic from mqtt, topic:$topic, message:$json" }
                }
            }
        }
    }

    private fun UMqttClient.setState(state: Int) {
        logger.info { "set state: $state" }
        setProperty(mapOf("state" to PropertyValue(state)))
    }

    private fun UMqttClient.setServerInfoWhenInitToFree() {
        val networkInfo = LocalNetInfo().getIpAndMacInfo().map { (mac, ip) ->
            mapOf("mac" to mac, "ip" to ip)
        }
        logger.info { "report device info to mqtt" }
        setProperty(
            mapOf(
                "storeId" to PropertyValue(appConfig.storeId.toInt()),
                "state" to PropertyValue(State.Free.id),
                "images" to PropertyValue(dockerClient.getContainerList()),
                "network" to PropertyValue(networkInfo)
            )  + DoorState.getDefaultForMqtt()
        )
    }
}
