package com.mjitech.service.state

import com.fasterxml.jackson.databind.node.ObjectNode
import com.mjitech.service.state.entity.Event
import com.mjitech.service.state.entity.SideEffect
import com.mjitech.service.state.entity.State
import com.mjitech.third.down.entity.server.HeartBeatResponse
import com.mjitech.third.down.entity.server.OutGoodStatus
import com.timzaak.ucloud.mqtt.entity.PropertyValue
import com.tinder.StateMachine
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.channels.BroadcastChannel
import kotlinx.coroutines.channels.ReceiveChannel
import kotlinx.coroutines.launch
import mu.KotlinLogging
import org.joda.time.DateTime
import kotlin.coroutines.CoroutineContext

class StateManager(override val coroutineContext: CoroutineContext) : CoroutineScope {
    private val logger = KotlinLogging.logger {}

    // Begin:WebResponseCache
    private val webResponseCache = WebResponseCache()

    fun setErrorOrderInfo(json: ObjectNode) {
        webResponseCache.setOrderInfo(json)
    }

    fun setWebNeedReload() {
        webResponseCache.setReload()
    }

    fun getWebResponse(screenSize: Int): WebResponseCache.CacheData {
        return webResponseCache.getCacheData(screenSize)
    }
    // End:WebResponseCache

    // Begin: DoorState
    private var doorState = DoorState()

    // TODO: PropertyValue 不应该放这里，但懒得写转换，就放这里了
    fun setNewDoorState(newDoorState: DoorState):Map<String, PropertyValue<Int>>? {
        val diffInfo = doorState.getDiffMapForMqtt(newDoorState)
        if (diffInfo.isEmpty()) {
            return null
        } else {
            doorState = newDoorState
            return diffInfo
        }
    }
    // End: DoorState


    val state: State
        get() = stateMachine.state

    fun handleEvent(event: Event) {
        stateMachine.transition(event)
    }

    // 下位机器状态, 为了兼容老逻辑, 没有融入到状态机里面去
    var downMachineState: HeartBeatResponse? = null

    fun setDownClientStatus(downHeartBeat: HeartBeatResponse) {
        this.downMachineState = downHeartBeat
    }

    private val channel: BroadcastChannel<StateTransition> = BroadcastChannel(1)

    fun subscribe(): ReceiveChannel<StateTransition> {
        return channel.openSubscription()
    }

    private val stateMachine = StateMachine.create<State, Event, SideEffect> {
        initialState(State.Init())

        state<State.Init> {
            // on <Any???>, 减少重复代码
            on<Event.InitEvent> { event ->
                val newState = if (event is Event.InitEvent.DownClientInitSuccess) {
                    this.copy(downClientInit = true)
                } else if (event is Event.InitEvent.MqttClientInitSuccess) {
                    this.copy(mqttClientInit = true)
                } else {
                    this
                }
                if (newState.isAllInit()) {
                    transitionTo(State.Free())
                } else {
                    transitionTo(newState)
                }
            }
        }

        state<State.Free> {
            onEnter {
                // reset every thing
                webResponseCache.reset()
            }
            on<Event.ScanCodeEvent> { event ->
                val step = 0
                val batchStep = 0
                transitionTo(
                    State.DoBusiness.TakingGood(event.order,batchStep, step, null),
                    SideEffect.TakeBatch(event.order,batchStep, step)
                )
            }
            on<Event.SetState> { event ->
                // Attention: 目前只相应 Free -> Repair 的变化
                if (event.stateId == State.Inspection.Repair.id) {
                    transitionTo(State.Inspection.Repair, SideEffect.EnterRepair)
                } else {
                    dontTransition()
                }
            }
            on<Event.OpenDoorInRepair> { event ->
                transitionTo(State.Inspection.OpenDoor(event.doorId), SideEffect.OpenDoorInRepair(event.doorId))
            }
        }

        state<State.DoBusiness.TakingGood> {
            on<Event.TakingSucc> {
                transitionTo(this.copy(),SideEffect.SetStatusTaking(order,batchStep,step))
            }
            on<Event.ProductRemain> {
                dontTransition()
            }
            on<Event.TakeToFree> {
                transitionTo(State.Free(DateTime.now()))
            }
            on<Event.DownPrepareForNext> {
                val step = 0
                val suSteps = succSteps + listOf(step)
                val batchStepNext = batchStep +1
                if(order.batches.size == batchStepNext){
                    //取完最后一个batch，置为free状态，修改订单状态
                    logger.info { "state to Free batchStep:$batchStepNext" }
                    transitionTo(State.Free(DateTime.now()),SideEffect.TakeBatch(order,batchStepNext, step))
                }else{
                    transitionTo(this.copy(step = step, batchStep = batchStepNext, succSteps = suSteps), SideEffect.TakeBatch(order,batchStepNext, step))
                }

            }
            on<Event.TakeGoodOk> {
                //取的货是否是当前batch中最后一个货
                if (order.batches[batchStep].skus.size - 1 == step) {
                    transitionTo(State.DoBusiness.OpenGoodDoor(order, batchStep,step), SideEffect.OpenDoor(order,batchStep))
                } else {
                    val nextStep = step + 1
                    val suSteps = succSteps + listOf(step)
                    transitionTo(this.copy(step = nextStep, succSteps = suSteps), SideEffect.TakeGood(order, batchStep, nextStep))
                }
            }
            on<Event.ChangeSkuFailure> {
                val nextStep = step+1
                if(nextStep == order.batches[batchStep].skus.size){
                    transitionTo(
                            State.DoBusiness.OpenGoodDoor(order,batchStep, nextStep),
                            SideEffect.OpenDoor(order,batchStep)
                    )
                }else{
                    transitionTo(this.copy(step=nextStep))
                }
            }
            on<Event.TakeGoodFailure> { event ->
                when (event.downStatus) {
                    OutGoodStatus.Companion.Status.TakeGoodMiss -> {
                        logger.info { "errorSteps : ${event.takingWhich}" }
                        // 存放出错步数
                        transitionTo(
                            this.copy(errorStep = event.takingWhich),
                            SideEffect.TakeGoodFailure(order, batchStep, event.takingWhich)
                        )
                    }
                    else -> {
                        dontTransition()
                    }
                }
                // transitionTo(State.MachineError.TakeGoodError(order.orderNumber, step, event.downStatus))
                // dontTransition()
            }
            on<Event.ChangeSku> { event ->
                transitionTo(this.copy(failTaking = event.failTaking), SideEffect.ChangeSku(event.order, batchStep, event.step))
            }
            on<Event.MachineError> { event ->
                logger.info { event.downStatus }
                when (event.downStatus) {
                    OutGoodStatus.Companion.Status.MachineError -> {
                        logger.info { "event MachineError ${event.downStatus}" }
                        transitionTo(
                            State.MachineError.TakeGoodError(order.orderNumber, step, event.downStatus.toInt()),
                            SideEffect.MachineError(order, batchStep, step + 1, step)
                        )
                    }
                    else -> {
                        dontTransition()
                    }
                }
                // transitionTo(State.MachineError.TakeGoodError(order.orderNumber, step, event.downStatus))
                // dontTransition()
            }
            on<Event.SetState> { event ->
                if (event.stateId == State.Free.id) {
                    transitionTo(State.Free(), SideEffect.BackToFree)
                } else {
                    dontTransition()
                }
            }
        }

        state<State.DoBusiness.OpenGoodDoor> {
            on<Event.OpeningDoor> {
                transitionTo(this.copy(isOpening = true))
            }
            on<Event.OpenSuccessDoor> {
                transitionTo(State.DoBusiness.TakingGood(order,batchStep,step,null))
            }
        }

        state<State.MachineError.TakeGoodError> {
            on<Event.UpdateMachinePc> {
                transitionTo(State.Inspection.UpdateMachinePc)
            }
        }

        state<State.Inspection> {
            on<Event.OpenDoorInRepair> { event ->
                if (this is State.Inspection.Repair) {
                    transitionTo(State.Inspection.OpenDoor(event.doorId), SideEffect.OpenDoorInRepair(event.doorId))
                } else if (this is State.Inspection.OpenDoor) {
                    if (this.isOpen(event.doorId)) {
                        if (event.isForce) {
                            dontTransition(SideEffect.OpenDoorInRepair(event.doorId))
                        } else {
                            dontTransition()
                        }
                    } else {
                        transitionTo(this.openDoor(event.doorId), SideEffect.OpenDoorInRepair(event.doorId))
                    }
                } else {
                    dontTransition()
                }
            }
        }

        state(
            StateMachine.Matcher.any<State, State>().where {
                !(this is State.MachineError.MachineIsNotSafe)
            }
        ) {
            on<Event.NewSkuPass> { event ->
                transitionTo(
                    State.Inspection.UpdateMachinePc, SideEffect.TestTakingSku(
                        event.shelfNumber,
                        event.x,
                        event.y,
                        event.x1,
                        event.y1
                    )
                )
            }
        }

        // 测试模拟状态改变
        state(StateMachine.Matcher.any<State, State>()) {
            on<Event.TestSetState> {event ->
                when (event.stateId) {
                    State.Free.id -> {
                        transitionTo(State.Free(),SideEffect.TestDownMachineLost(false))
                    }
                    State.MachineError.DownMachineLost.id -> {
                        transitionTo(State.MachineError.DownMachineLost,SideEffect.TestDownMachineLost(true))
                    }
                    State.DoBusiness.TakingGood.id -> {
                        transitionTo(State.DoBusiness.TakingGood(event.order,0,0,null))
                    }
                    State.MachineError.TakeGoodError.id -> {
                        transitionTo(State.DoBusiness.TakingGood(event.order,0,0,null, emptyList(),1))
                    }
                    State.DoBusiness.OpenGoodDoor.id -> {
                        transitionTo(State.DoBusiness.OpenGoodDoor(event.order,0,0,true))
                    }
                    State.Inspection.Update.id -> {
                        transitionTo(State.Inspection.UpdateMachinePc)
                    }
                    else -> {
                        dontTransition()
                    }
                }

            }
        }

        // 异常处理机制
        state(
            StateMachine.Matcher.any<State, State>().where {
                !(this is State.MachineError)
            }
        ) {
            on<Event.DownMachineLost> {
                transitionTo(State.MachineError.DownMachineLost)
            }

            on<Event.DownMachineIsNotSafe> { event ->
                transitionTo(
                    State.MachineError.MachineIsNotSafe(
                        event.leftHorseDoor,
                        event.rightHorseDoor,
                        event.replenishmentScreen
                    )
                )
            }
        }

        // 异常恢复
        state(StateMachine.Matcher.any<State, State.MachineError>()) {
            on<Event.SetState> { event ->
                // TODO: 是否要强制,才能恢复到 Free
                if (event.stateId == State.Free.id) {
                    transitionTo(State.Free(), SideEffect.BackToFree)
                } else if (event.stateId == State.Inspection.Repair.id) {
                    transitionTo(State.Inspection.Repair, SideEffect.EnterRepair)
                } else {
                    dontTransition()
                }
            }
        }

        onTransition { transition ->
            logger.debug { "come event ${transition.event}, init state: ${transition.fromState}" }
            val validTransition = transition as? StateMachine.Transition.Valid ?: return@onTransition
            if (validTransition.fromState != validTransition.toState) {
                logger.info { "transition: ${validTransition.fromState} -> ${validTransition.toState}" }
            }
            launch {
                channel.send(validTransition)
            }

        }
    }
}
typealias StateTransition = StateMachine.Transition.Valid<State, Event, SideEffect>
