package com.mjitech.service.state.entity

import com.mjitech.api.model.FailTaking
import com.mjitech.iot.DeviceStatus
import com.mjitech.iot.DoorId
import com.mjitech.api.model.Order
import org.joda.time.DateTime

interface StateDesc {
    val id: Int
}

sealed class State {
    abstract fun desc(): StateDesc

    data class Init(val downClientInit:Boolean= false, val mqttClientInit:Boolean = false) : State() {
        companion object : StateDesc {
            override val id: Int = DeviceStatus.Init
        }

        fun isAllInit():Boolean {
            return downClientInit && mqttClientInit
        }
        override fun desc(): StateDesc {
            return Companion
        }
    }

    data class Free(val heartBeatTime: DateTime = DateTime.now()) : State() {
        companion object : StateDesc {
            override val id: Int = DeviceStatus.Free
        }

        override fun desc(): StateDesc {
            return Companion
        }
    }

    sealed class DoBusiness : State() {
        data class TakingGood(
            val order: Order,
            val batchStep: Int,
            val step: Int,
            val failTaking: FailTaking?,
            val succSteps: List<Int> = emptyList(),
            val errorStep: Int = -1
        ) : DoBusiness() {
            companion object : StateDesc {
                override val id: Int = DeviceStatus.TakingGood
            }

            override fun desc(): StateDesc {
                return Companion
            }
        }

        data class OpenGoodDoor(val order: Order,val batchStep: Int, val step: Int, val isOpening: Boolean = false, val errorSteps: List<Int> = emptyList()) : DoBusiness() {
            companion object : StateDesc {
                override val id: Int = DeviceStatus.OpenGoodDoor
            }

            override fun desc(): StateDesc {
                return Companion
            }
        }
    }

    // 维护状态
    sealed class Inspection : State() {
        // 进入维护状态
        object Repair : Inspection(), StateDesc {
            override val id: Int = DeviceStatus.InRepair

            override fun desc(): StateDesc {
                return this
            }
        }

        // 补货
        data class OpenDoor(val door: Int) : Inspection() {
            companion object : StateDesc {
                val ElectronDoor = DoorId.ElectronDoor
                val ADoor = DoorId.ADoor
                val BDoor = DoorId.BDoor
                val OutGoodDoor = DoorId.OutGoodDoor
                override val id: Int = DeviceStatus.OpenDoorInRepair
            }

            override fun desc(): StateDesc {
                return Companion
            }

            fun openDoor(doorId: Int): OpenDoor {
                assert(doorId == OutGoodDoor || doorId == ElectronDoor || doorId == ADoor || doorId == BDoor)
                return this.copy(door or doorId)
            }

            fun isOpen(doorId: Int): Boolean {
                return door and doorId == doorId
            }
        }

        // 升级
        object Update : Inspection(), StateDesc {
            override val id: Int = DeviceStatus.UpdateInReapire
            override fun desc(): StateDesc {
                return this
            }
        }
        //
        object UpdateMachinePc : Inspection(), StateDesc {
            override fun desc(): StateDesc {
                return this
            }
            override val id: Int = DeviceStatus.UpdateMachinePc
        }

    }

    sealed class MachineError : State() {
        data class TakeGoodError(val orderNumber: String, val step: Int, val downErrorCode: Int) : MachineError() {
            companion object : StateDesc {
                override val id: Int = DeviceStatus.TakeGoodError
            }

            override fun desc(): StateDesc {
                return Companion
            }
        }

        object DownMachineLost : MachineError(), StateDesc {
            override fun desc(): StateDesc {
                return this
            }

            override val id: Int = DeviceStatus.DownMachineLost
            override fun toString(): String {
                return "State[DownMachineLost]"
            }
        }

        data class MachineIsNotSafe(
            val leftHorseDoor: Byte,
            val rightHorseDoor: Byte,
            val replenishmentScreen: Byte
        ) : MachineError() {
            companion object : StateDesc {
                override val id: Int = DeviceStatus.MachineIsNotSafe
            }

            override fun desc(): StateDesc {
                return Companion
            }
        }
    }
}
