package com.timzaak.ucloud.mqtt

import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.databind.node.ObjectNode
import com.timzaak.ucloud.mqtt.entity.MqttState
import com.timzaak.ucloud.mqtt.entity.PostEvent
import com.timzaak.ucloud.mqtt.entity.PostProperty
import com.timzaak.ucloud.mqtt.entity.PropertyValue
import com.timzaak.ucloud.mqtt.entity.Request
import com.timzaak.ucloud.mqtt.entity.ResponseCommand
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import mu.KotlinLogging
import org.eclipse.paho.client.mqttv3.IMqttDeliveryToken
import org.eclipse.paho.client.mqttv3.MqttCallback
import org.eclipse.paho.client.mqttv3.MqttClient
import org.eclipse.paho.client.mqttv3.MqttConnectOptions
import org.eclipse.paho.client.mqttv3.MqttMessage
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence
import kotlin.coroutines.CoroutineContext

// UCloud MQTT Client WRAPPER
class UMqttClient(
    private val conf: MqttConfig,
    private val objectMapper: ObjectMapper,
    //TODO: make channel private
    val mqttMessageChannel: MqttMessageChannel,
    val mqttStateChannel: MqttStateChannel,
    override val coroutineContext: CoroutineContext
) : MqttCallback, CoroutineScope {
    private val logger = KotlinLogging.logger {}
    val persistence: MemoryPersistence = MemoryPersistence()
    var client: MqttClient = MqttClient(conf.broker, conf.clientId, persistence)

    init {
        connect()
    }

    private fun connect() {
        // persistence.clear()
        client = MqttClient(conf.broker, conf.clientId, persistence)
        val connOpt = MqttConnectOptions()
        connOpt.password = conf.devicePassword.toCharArray()
        connOpt.userName = conf.userName
        connOpt.isCleanSession = true
        client.setCallback(this)

        client.connect(connOpt)
        logger.info { "MQTT Client try to connect" }
        launch {
            var tryNum = 1
            while (tryNum < 3) {
                delay((1000 * tryNum).toLong())
                tryNum += 1
                if (client.isConnected) {
                    subscribe(CommandTopic)
                    logger.info { "subscribe UCloud command topic" }
                    subscribe("/\$system${conf.topicPrefixPath}tmodel/property/post_reply")
                    logger.info { "subscribe UCloud set property reply topic" }
                    mqttStateChannel.send(MqttState.Connected)
                    break
                } else {
                    logger.warn("MQTT Client does not connect")
                    if (tryNum == 3) {
                        logger.warn("MQTT Client subscribe failure")
                    }
                }
            }
        }
    }

    private fun reConnect() {
        client.close()
        logger.info { "MQTT Client try to reconnect" }
        connect()
        logger.info { "MQTT Client reconnect" }
    }

    fun publish(topic: String, json: ByteArray) {
        if (client.isConnected) {
            val message = MqttMessage(json)
            message.qos = conf.qos
            try {
                client.publish(topic, message)
            } catch (e: Throwable) {
                logger.error(e) { "send message error" }
            }
        } else {
            logger.warn("send message before client connect, topic:$topic, data:${json}")
        }
    }

    private fun publishRequest(topic: String, value: Request) {
        val data = objectMapper.writeValueAsBytes(value)
        if (client.isConnected) {
            logger.debug { "publish request: $topic, data: ${objectMapper.writeValueAsString(value)}" }
            publish(topic, data)
        } else {
            logger.warn { "mqtt lost connection, not send request: $topic, data:${objectMapper.writeValueAsString(value)}" }
        }
    }

    fun subscribe(topicFilter: String) {
        kotlin.runCatching {
            client.subscribe(topicFilter)
        }.onFailure { exception ->
            logger.error(exception) { "subscribe $topicFilter fail" }
        }
    }

    fun disconnect() {
        client.disconnect()
    }

    override fun messageArrived(topic: String, message: MqttMessage) {
        launch {
            mqttMessageChannel.send(topic to message)
        }
    }

    override fun connectionLost(cause: Throwable) {
        // TODO: 优化重试策略
        launch {
            delay(5000)
            reConnect()
        }
    }

    override fun deliveryComplete(token: IMqttDeliveryToken) {
        // TODO("Not yet implemented")
    }

    fun setProperty(properties: Map<String, PropertyValue<*>>) {
        publishRequest("/\$system${conf.topicPrefixPath}tmodel/property/post", PostProperty(properties))
    }

    fun postEvent(identifier: String, event: ObjectNode) {
        publishRequest("/\$system${conf.topicPrefixPath}tmodel/event/post", PostEvent(identifier, event))
    }

    fun respCommand(resp: ResponseCommand) {
        publishRequest("/\$system${conf.topicPrefixPath}tmodel/command_reply/${resp.RequestID}", resp)
    }

    //    val PostEventTopic by lazy {
//        "${conf.topicPrefixPath}tmodel/event/post"
//    }
//    val SetPropertyTopic by lazy { "${conf.topicPrefixPath}tmodel/property/post" }
    val CommandTopic by lazy {
        "/\$system${conf.topicPrefixPath}tmodel/command"
    }
}
