package helper

import com.mjitech.api.BaseAPIImpl
import com.mjitech.api.OrderAPI
import com.mjitech.iot.IotClient
import com.mjitech.service.ShopService
import com.mjitech.service.state.StateManager
import com.mjitech.third.down.DownClient
import com.mjitech.third.down.DownClientChannel
import com.mjitech.third.down.DownUDPClient
import com.mjitech.third.down.mock.DownServerHeartBeat
import com.mjitech.third.down.mock.MockDownServer
import com.mjitech.third.down.mock.MockUDPDownServer
import com.mjitech.third.scancode.ScanCodeChannel
import com.mjitech.third.scancode.ScanCodeClient
import com.mjitech.third.scancode.mock.MockScanCodeServer
import com.timzaak.coroutines.channel.BaseScopeBroadcastChannel
import com.timzaak.coroutines.channel.BaseScopeChannel
import com.timzaak.jackson.KtorJacksonWrapper
import com.timzaak.ucloud.iot.UIotServerClient
import com.timzaak.ucloud.mqtt.MqttMessageChannel
import com.timzaak.ucloud.mqtt.MqttConfig
import com.timzaak.ucloud.mqtt.UMqttClient
import com.typesafe.config.ConfigFactory
import config.AppConfig
import config.isEnableMockScanCodeServer
import di.apiDI
import di.thirdDI
import io.ktor.client.HttpClient
import io.ktor.client.engine.apache.Apache
import io.ktor.client.features.logging.DEFAULT
import io.ktor.client.features.logging.LogLevel
import io.ktor.client.features.logging.Logger
import io.ktor.client.features.logging.Logging
import kotlinx.coroutines.GlobalScope.coroutineContext
import kotlinx.coroutines.newSingleThreadContext
import mu.KotlinLogging
import org.koin.dsl.module
import org.koin.test.KoinTestRule
import java.net.InetSocketAddress
import kotlin.coroutines.CoroutineContext

private val logger = KotlinLogging.logger { }
private val config = ConfigHelper.config

fun scanCodeClientRule(): KoinTestRule {
    return KoinTestRule.create {
        modules(
            module {
                single {
                    newSingleThreadContext("test scanCode") as CoroutineContext
                }
                single(createdAtStart = true) {
                    ScanCodeChannel(BaseScopeChannel(get()))
                }

                single(createdAtStart = true) {
                    ScanCodeClient(config.getString("third.scanCodeClient.port"), get())
                }

                if (config.isEnableMockScanCodeServer()) {
                    single(createdAtStart = true) {
                        MockScanCodeServer(config.getString("mock.third.scanCodeServer.port"))
                    }
                }
            }
        )
    }
}

fun shopServiceDI(): KoinTestRule {
    return KoinTestRule.create {
        logger.debug("init service rule")
        val config = ConfigFactory.load()
        val appConfig = AppConfig(config)
        val httpClient = HttpClient(Apache) {
            install(Logging) {
                logger = Logger.DEFAULT
                level = LogLevel.ALL
            }
        }
        val baseAPIImpl = BaseAPIImpl(
            appConfig, httpClient, coroutineContext
        )
        modules(
            thirdDI(config),
            apiDI(),
            module {
                single { config }
//                single { coroutineContext }
                single { appConfig }
                single { httpClient }
                single { baseAPIImpl }
                single {
                    newSingleThreadContext("test service") as CoroutineContext
                }
                single {
                    StateManager(get())
                }
                single {
                    KtorJacksonWrapper.jacksonMapper
                }
                single {
                    ShopService(get(), get(), get(), get(), get(), get(), get())
                }
            }
        )
    }
}

fun mockDownRule(): KoinTestRule {
    return KoinTestRule.create {
        logger.debug("init Mock port rule")
        modules(
                module {
                    single {
                        newSingleThreadContext("test mock port") as CoroutineContext
                    }
                    single {
                        DownClientChannel(BaseScopeBroadcastChannel(1, get()))
                    }
                    single(createdAtStart = true) {
                        DownUDPClient(
                                InetSocketAddress(
                                        config.getString("third.downClient.downServer.ip"),
                                        config.getInt("third.downClient.downServer.port")
                                ),
                                get(),
                                get()
                        ) as DownClient
                    }
                    if (config.getBoolean("mock.third.downServer.enable")) {
                        single(createdAtStart = true) {
                            MockUDPDownServer(
                                    config.getInt("third.downClient.downServer.port"),
                                    get()
                            )
                        }
                    }
                }
        )
    }
}

fun downClientRule(): KoinTestRule {
    return KoinTestRule.create {
        logger.debug("init serial port rule")
        modules(
            module {
                single {
                    newSingleThreadContext("test serail port") as CoroutineContext
                }
                single {
                    DownClientChannel(BaseScopeBroadcastChannel(1, get()))
                }
                single(createdAtStart = true) {
                    DownUDPClient(
                        InetSocketAddress(
                            config.getString("third.downClient.downServer.ip"),
                            config.getInt("third.downClient.downServer.port")
                        ),
                        get(),
                        get()
                    ) as DownClient
                }
                if (config.getBoolean("mock.third.downServer.enable")) {
                    single(createdAtStart = true) {
                        MockUDPDownServer(
                            config.getInt("third.downClient.downServer.port"),
                            get()
                        ) as MockDownServer
                    }
                    single(createdAtStart = true) {
                        DownServerHeartBeat(get(), get())
                    }
                }
            }
        )
    }
}

fun downUdpRule(): KoinTestRule {
    return KoinTestRule.create {
        logger.debug("init serial port rule")
        modules(
                module {
                    single {
                        newSingleThreadContext("test serail port") as CoroutineContext
                    }
                    single {
                        DownClientChannel(BaseScopeBroadcastChannel(1, get()))
                    }
                    single(createdAtStart = true) {
                        DownUDPClient(
                                InetSocketAddress(
                                        config.getString("third.downClient.downServer.ip"),
                                        config.getInt("third.downClient.downServer.port")
                                ),
                                get(),
                                get()
                        ) as DownClient
                    }
                }
        )
    }
}

fun apiRule(): KoinTestRule {
    return KoinTestRule.create {
        logger.debug("init api port rule")
        modules(
            module {
                single {
                    BaseAPIImpl(
                        AppConfig(config), HttpClient(Apache), coroutineContext
                    )
                }
                single {
                    OrderAPI(get())
                }
            }
        )
    }
}

fun mqttRule(): KoinTestRule {
    return KoinTestRule.create {
        logger.debug("init mqtt")
        modules(
            module {
                single {
                    newSingleThreadContext("test serail port") as CoroutineContext
                }
                single {
                    config
                }
                single {
                    MqttMessageChannel(BaseScopeBroadcastChannel(1, get()))
                }
                single {
                    KtorJacksonWrapper.jacksonMapper
                }
                single {
                    UMqttClient(MqttConfig.fromConfig(config.getConfig("api.mqtt")), get(), get(), get(),get())
                }
                single {
                    val c = config.getConfig("api.mqttServer")
                    UIotServerClient(
                        c.getString("privateKey"),
                        c.getString("publicKey"),
                        config.getString("api.mqtt.productSN")
                    )
                }
                single {
                    IotClient(get(), get())
                }
            }
        )
    }
}
